"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const { declare } = require('@babel/helper-plugin-utils');
const core_1 = require("@babel/core");
exports.default = declare((api) => {
    api.assertVersion(7);
    const LINGUI_MODULE = '@lingui/react';
    const REACT_INTL_MODULE = 'react-intl';
    const moduleMap = new Map([[REACT_INTL_MODULE, LINGUI_MODULE]]);
    const componentMap = new Map([['FormattedMessage', 'Trans'], ['FormattedHTMLMessage', 'Trans']]);
    const attributeMap = new Map([['defaultMessage', 'defaults']]);
    function isTranslationImport(path) {
        return path.isImportDeclaration() && moduleMap.has(path.node.source.value);
    }
    function replaceTranslationModule(importPath, file) {
        const unsupportedImports = [];
        const addedNames = new Map();
        const specifiers = importPath.get('specifiers').reduce((list, elem) => {
            if (elem.isImportSpecifier()) {
                const name = componentMap.get(elem.node.imported.name);
                if (name) {
                    if (!addedNames.has(name)) {
                        addedNames.set(name, elem);
                        list.push(core_1.types.importSpecifier(core_1.types.identifier(name), core_1.types.identifier(name)));
                    }
                }
                else {
                    unsupportedImports.push(elem.node.imported.name);
                }
            }
            else {
                throw Error('Just only support named import');
            }
            return list;
        }, []);
        if (unsupportedImports.length) {
            console.warn('Skipping this file due to still using unsupported methods: %s at line %s > %s', unsupportedImports.join(','), importPath.node.source.start, file.opts.filename);
            return false;
        }
        importPath.node.specifiers = specifiers;
        // importPath.get('specifiers').forEach(elem => {
        //   elem.scope.registerBinding('module', elem);
        // });
        importPath.get('source').replaceWith(core_1.types.stringLiteral(LINGUI_MODULE));
        return true;
    }
    function isTranslationComponent(path) {
        const name = path.get('name');
        if (name.isJSXMemberExpression()) {
            const object = name.get('object');
            const binding = object.scope.getBinding(object.node.name);
            if (binding) {
                const parentBinding = binding.path.parentPath;
                if (parentBinding.isImportDeclaration() && parentBinding.get('source').node.value === REACT_INTL_MODULE) {
                    return true;
                }
            }
            return false;
        }
        return core_1.types.isJSXIdentifier(path.node.name) && isReferencedToReactIntlModule(path.get('name'));
    }
    function isReferencedToReactIntlModule(path) {
        // console.log(path.referencesImport(LINGUI_MODULE, 'FormattedMessage'));
        return Array.from(componentMap.keys()).some(k => path.referencesImport(LINGUI_MODULE, k));
    }
    function replaceTranslationJSXAttributes(jsxOpeningElementPath) {
        const processAttributes = () => {
            const replaceValuePairKeys = new Map([
                ['id', 'defaultMessage'],
            ]);
            const setKeys = new Set([
                ...replaceValuePairKeys.keys(),
                ...replaceValuePairKeys.values(),
            ]);
            const db = new Map();
            jsxOpeningElementPath.get('attributes').forEach(item => {
                if (item.isJSXAttribute()) {
                    const name = item.get('name');
                    if (name.isJSXIdentifier()) {
                        if (setKeys.has(name.node.name)) {
                            db.set(name.node.name, item);
                        }
                        if (attributeMap.has(name.node.name)) {
                            item.node.name = core_1.types.jsxIdentifier(attributeMap.get(name.node.name));
                        }
                    }
                }
            });
            for (let [k, v] of replaceValuePairKeys) {
                const key = db.get(k);
                const value = db.get(v);
                if (key && value) {
                    key.get('value').replaceWith(value.get('value'));
                }
            }
        };
        if (core_1.types.isJSXIdentifier(jsxOpeningElementPath.node.name)) {
            const linguiName = jsxOpeningElementPath.node.name.name;
            jsxOpeningElementPath.node.name.name = componentMap.get(linguiName);
            processAttributes();
        }
        else {
            const name = jsxOpeningElementPath.get('name');
            if (name.isJSXMemberExpression()) {
                const object = name.get('object');
                const property = name.get('property');
                object.node.name = 'react-intl';
                property.node.name = componentMap.get(property.node.name);
                processAttributes();
            }
        }
    }
    let hasTrans;
    let didExit;
    return {
        visitor: {
            Program: {
                enter(path, { file }) {
                    const lingImports = [];
                    const intlImports = [];
                    path.get('body').forEach(statement => {
                        if (statement.isImportDeclaration()) {
                            if (isTranslationImport(statement)) {
                                hasTrans = true;
                                const status = replaceTranslationModule(statement, file);
                                if (status) {
                                    intlImports.push(statement);
                                }
                            }
                            else {
                                if (statement.node.source.value === LINGUI_MODULE) {
                                    lingImports.push(statement);
                                }
                            }
                        }
                    });
                    if (intlImports.length > 1) {
                        throw Error(`This plugin currently don't support multiple import from "react-intl" for now`);
                    }
                    // both FormattedMessage and Trans get imported
                    // so we remove one
                    if (lingImports.length > 0 && intlImports.length === 1) {
                        const linguiSpecifiers = [];
                        lingImports[0].get('specifiers').forEach(lSpecifier => {
                            intlImports.forEach(intlImport => {
                                const duplicated = intlImport.get('specifiers').some(rSpecifier => rSpecifier.node.local.name === lSpecifier.node.local.name);
                                if (!duplicated) {
                                    linguiSpecifiers.push(lSpecifier);
                                }
                            });
                        });
                        intlImports[0].get('specifiers').concat(linguiSpecifiers);
                        for (const item of lingImports) {
                            item.remove();
                        }
                    }
                },
                exit() {
                    didExit = true;
                }
            },
            JSXElement(jsxElementPath) {
                if (!hasTrans) {
                    return;
                }
                const jsxOpenElementImportPath = jsxElementPath.get('openingElement');
                if (isTranslationComponent(jsxOpenElementImportPath)) {
                    replaceTranslationJSXAttributes(jsxOpenElementImportPath);
                }
            },
        },
    };
});
